// Code with Encoder:
// Test MD03a / Pololu motor with encoder
// http://forum.arduino.cc/index.php?topic=8652.0
// speed control (PI), V & I display
// Credits:
//   Dallaby   http://letsmakerobots.com/node/19558#comment-49685
//   Bill Porter  http://www.billporter.info/?p=286
//   bobbyorr (nice connection diagram) http://forum.pololu.com/viewtopic.php?f=15&t=1923
//   pin changes by ewh for the VNH5019
//   Changed pinb reading to read pin 11, should be BIT 4, was BIT 3, encodepin b - 4 Nov 2014
//   motor and chip example - Manual pid calc
//   http://forum.arduino.cc/index.php?topic=8652.0  the basic site
//   http://forum.arduino.cc/index.php/topic,8652.0.html
//   http://www.pololu.com/product/708  the driver chip
//   http://www.pololu.com/catalog/product/1443 the motor
/************************************************
REVIEW PIN to ARDUINO SETTINGS FOR BOTH DRIVERS!! 
Check code for pre-defined pins for both drivers
v-v-v-v-v-v-v-
4 Nov 2014 - ewh - Added the pin change interuppt vector code
                   this works for speed change, etc.
5 Nov 2014 - ewh - Added code for direct pin setting
7 Nov 2014 - ewh - Revising PID settings and current variable
12 Nov 2014 - ewh - Added pid autotune code (sucks)
17 Nov 2014 - ewh - More pin changes, switched from 11,12, to 
                    11 and 13! change interuppte to PCINT5!!
18 Nov 2014 - ewh - added code for reverse direction
************************************************************/

/*------------------------------------------------ *
Truth Table for InA1 and InB1
INa	INb
1	1	Brake to Vcc
1	0	Clockwise Rotation
0	1	Counterclockwise Rotation
0	0 	Brake to GND
*-------------------------------------------------*/

//***#include <DualVNH5019MotorShield.h>
#include <PID_v1.h>
#include <Wire.h>
//#include <PID_AutoTune_v0.h>

//***DualVNH5019MotorShield md; // default pins ok for this

//define variables we will be connecting to
//double Setpoint, Input, Output;
//PID myPID(&Input, &Output, &Setpoint, 0.4,1,1, DIRECT);


#define InA1            2       // INA motor pin
#define InB1            4       // INB motor pin 
#define PWM1            9       // PWM motor pin			       		
#define encodPinA1      11      // encoder A pin
#define encodPinB1      13      // encoder B pin
#define Vpin            A3      // battery monitoring analog pin
#define Apin            A0      // motor current monitoring analog pin

#define CURRENT_LIMIT   3000                    // high current warning
#define LOW_BAT         8000 //10000            // low bat warning
#define LOOPTIME        100                     // PID loop time
#define NUMREADINGS     10                      // samples for Amp average
#define DIRECT   0
#define REVERSE  1

int readings[NUMREADINGS];
//int readings[] = {0,0,0,0,0,0,0,0,0,0};
unsigned long lastMilli = 0;                    // loop timing 
unsigned long lastMilliPrint = 0;               // loop timing
int speed_req = 40;                             // speed (Set Point)  
int speed_act = 0;                              // speed (actual value)
int PWM_val = 0;                                // (25% = 64; 50% = 127; 75% = 191; 100% = 255)
int voltage = 0;                                // in mV
int current = 0;                                // in mA
int controllerDirection;
volatile long count = 0;                        // rev counter
float Kpi = 0.3;//.4;                           // PID proportional control Gain
float Kdi =  2;                                 // PID Derivitave control gain

byte ATuneModeRemember=2;
double Input, Output, Setpoint; //180
double kp=0.3,ki=1,kd=0; //2,.5,2

double kpmodel=1.5, taup=100, theta[50];

unsigned long  modelTime, serialTime;

PID myPID(&Input, &Output, &Setpoint,kp,ki,kd, DIRECT);

//PID_ATune aTune(&input, &output);
//PID myPID(&Input, &Output, &Setpoint, 0.4,1,1, REVERSE);
//set to false to connect to the real world
//boolean useSimulation = false;//true;

void setup() {
  analogReference(EXTERNAL);  // Current external ref is 3.3V, appiled to the aREF pin
  Serial.begin(115200);
  pinMode(InA1, OUTPUT);
  pinMode(InB1, OUTPUT);
  pinMode(PWM1, OUTPUT);
  cli();  // disable interrupts 
  // CANNOT USE ARDUINO PIN 8,9,10,12,A0,A1 for non-controller/Arduino signals
  // bit-pin map 0=8,1=9,2=10,3=11,4=12,5=13 "B" register (digital pin 8 to 13)
  // cannot use bits 6 and 7 on reg B
    DDRB  |= (0 << 3);  //  Enable pin 11 as INPUT for encodPinA1 set low (was 2)
    PORTB |= (1 << 3);  //  Set Arduino pin 11 HIGH for pull up (encodPinA1) (was 2)
    DDRB  |= (0 << 5);  //  Enable pin 13 as INPUT for limit switch - set low (was 3)
    PORTB |= (1 << 5);  //  Set Arduino pin 13 HIGH for pull up and encodPinB1 (was 3)
    DDRD  |= (0 << 2);  //  Set direction for Arduino pin 2 as input
    PORTD |= (1 << 2);  //  Set digial pin 2 HIGH CW
    DDRD  |= (0 << 4);  //  Set direction for Arduino pin 4 as input
    PORTD |= (0 << 4);  //  Set digial pin 4 LOW  CW  
    //digitalWrite(InA1, LOW);  //low  pid has trouble with direction
    //digitalWrite(InB1, HIGH); //high   ccw, H,L=cw  
    //pinMode(encodPinA1, INPUT); 
    //pinMode(encodPinB1, INPUT); 
    //digitalWrite(encodPinA1, HIGH);      // turns on pullup resistor
    //digitalWrite(encodPinB1, HIGH);    
    PCICR  |= (1 << PCIE0);   // enable group interrupts on PORTB PCINT[7:0]
    PCMSK0 |= (1 << PCINT5); // enable interrupt pin 11 aka bit 3, try 10 for B, 11 forA
  sei();  // enable interrupts  bit 3=11 bit 2 = 10
  //attachInterrupt(1, rencoder, FALLING); 
   for(int i=0; i<NUMREADINGS; i++) readings[i] = 0;  // initialize readings to 0    
   analogWrite(PWM1, PWM_val);  
   Input = speed_act;
   speed_req = 40;
   Setpoint = speed_req;
   myPID.SetOutputLimits(-255, 255);
  //turn the PID on
   myPID.SetMode(AUTOMATIC);  
   controllerDirection = DIRECT;
}

//ISR for Pin 8 was 0 and 1, change to 3 and 2 for pins 11 and 10
ISR (PCINT0_vect)    // handle pin change interrupt for D8-D13;  triggered by arduino pin 13
{	
if(bitRead(PINB,5)== bitRead(PINB,3)) {  //(PINB, x) is the syntax to read the input state from PORTB, bit x 
    //encoderValue++;    
    count++;     
  } else {
    //encoderValue--;
    count--;    
  } 
}
/*  Reverse Code Example 

void setup()
{
 ...
myPID.SetOutputLimits (-255,255); //instead of 0-255, allow the output to range -255 to 255;
...
}

void loop()
{
...
myPID.Compute()
//turn the output into direction and speed
if(Output<0)
{
  direction = backward; //don't know exactly how you set this in your situation
  speed = -1*Output
}
else
{
  direction = forward;
  speed = Output;
}
...
}

*/

void changeDIR(int CCW)
{
  switch (CCW) {
  	 case REVERSE:             
            speed_act = -1*speed_act;
            //kp = (0-kp);
            //ki = (0-ki);
            //kd = 0;
            //Output = -1*Output;
  	    digitalWrite(InA1, LOW);  // CCW
            digitalWrite(InB1, HIGH); // CCW	
            //myPID.SetControllerDirection(REVERSE);
  	break;
  	  case DIRECT:  	  
  	    //kp = fabs(kp);
  	    //ki = fabs(ki);
  	    //kd = fabs(kd);  
  	    //speed_req = -1*speed_req;  	   
  	    //speed_req = abs(speed_req);
  	    //Input = speed_act;
  	    digitalWrite(InA1, HIGH);  // CW
            digitalWrite(InB1, LOW);   // CW
            //myPID.SetControllerDirection(DIRECT);
  	  break;
  	 default:
  	 Serial.print("hello");
   }
}

void loop() {  
  
  getParam();                  // check keyboard
  if((millis()-lastMilli) >= LOOPTIME) {           // enter timed loop
      lastMilli = millis(); 
      getMotorData();                   // calculate speed, volts and Amps
 //   PWM_val= updatePid(PWM_val, speed_req, speed_act); // compute PWM value    
      Input = speed_act;    
      myPID.Compute();
      // or add this here if(controllerDirection == REVERSE) Output = -1*Output;
      analogWrite(PWM1, Output); //PWM_val);   // send PWM to motor    
      if(millis()>serialTime)
	{
	  SerialReceive();
	  SerialSend();
	  serialTime+=500; 
        }
   }   
      //if(millis()>serialTime)
      //  {
      //    SerialReceive();
      //    SerialSend();
      //    serialTime+=500; 
      //  }
    //printMotorInfo();     
}

void getMotorData()     // calculate speed, volts and Amps
{                                         
  static long countAnt = 0;  // last count // 64 pulses X 131 gear ratio = 8384, try 32 > 4192  
  speed_act = ((count - countAnt)*(60*(1000/LOOPTIME)))/(32*131);
  //if(controllerDirection == REVERSE) {
  //Serial.print("encoder: "); 
  //Serial.println(count);
  //speed_act = ((count - countAnt)*(60*(1000/LOOPTIME)))/(32*131); // 16 pulses X 29 gear ratio = 464 counts per output shaft rev
 //} 
  countAnt = count;   
  voltage = int(analogRead(Vpin) * 3.22 * 12.2/2.2);     // battery voltage: mV=ADC*3300/1024, voltage divider 10K+2K, was 12.2/2.2 for voltage
  //current = int(analogRead(Apin) * 3.22 * .77 *(1000.0/144.0));  // motor current - output: 130mV per Amp  132.0
  current = int(analogRead(Apin)*34);  //md.getM1CurrentMilliamps();
  //5.0 * .77 *(1000.0/144.0));analogRead(Apin) * 34;
  //// 5V / 1024 ADC counts / 144 mV per A = 34 mA per count
  current = digital_smooth(current, readings);                   // remove signal noise
}

void printProgData(int val1, int val2, int val3)
{
  Serial.print("command: ");  Serial.print(val1);
  Serial.print(" targetValue: "); Serial.print(val2);
  Serial.print(" currentValue: "); Serial.println(val2);
}

int updatePid(int command, int targetValue, int currentValue)     // compute PWM value
{    
  float pidTerm = 0;                                              // PID correction
  int error=0;                                  
  static int last_error=0;  
  Setpoint = targetValue;
  error = abs(targetValue) - abs(currentValue); 
  //printProgData(command,targetValue,currentValue);
  pidTerm = (Kpi * error) + (Kdi * (error - last_error));
  //Serial.print("PWM_val: ");
  //Serial.println(constrain(command + int(pidTerm), 0, 255));
  last_error = error;  
  return constrain(command + int(pidTerm), 0, 255);
}

void printMotorInfo()  {                      // display data
  
  if((millis()-lastMilliPrint) >= 500)   {                     
    lastMilliPrint = millis();
    Serial.print("SP:");             Serial.print(speed_req);  
    Serial.print("  RPM:");          Serial.print(speed_act);
    Serial.print("  PWM:");          Serial.print(PWM_val);  
    Serial.print("  V:");            Serial.print(float(voltage)/1000,1);
    Serial.print("  mA:");           Serial.println(current);

    if (current > CURRENT_LIMIT)            Serial.println("*** CURRENT_LIMIT ***");                
    if (voltage > 1000 && voltage*11 < LOW_BAT)   Serial.println("*** LOW_BAT ***");                
  }
}

int getParam()  {
  char param, cmd;
  if(!Serial.available())    return 0;
  delay(10);                  
  param = Serial.read();            // get parameter byte
  if(!Serial.available())    return 0;
  cmd = Serial.read();              // get command byte
  Serial.flush();
  
  switch (param) {
    case 'v':                       // adjust speed
      if(cmd=='+')  {
        speed_req += 10;       
        Setpoint = speed_req;
        if(speed_req>400) speed_req=400;
      }
      if(cmd=='-')    {
        if (speed_req< 24) {
              speed_req -= 1;
              }else{
              speed_req -= 10;
              }        
        Setpoint = speed_req;        
        if(speed_req<0)  speed_req=0;       
      }
      break;
    case 's':                     // adjust direction
      if(cmd=='+'){		
	      if (controllerDirection == REVERSE) {	 
            changeDIR(DIRECT);
            controllerDirection = DIRECT;
            Output = 1*Output;
	    }	   
        // digitalWrite(InA1, HIGH);  // CW
        // digitalWrite(InB1, LOW);   // CW
      }
      if(cmd=='-')   {
	   if (controllerDirection == DIRECT) {
	    	 changeDIR(REVERSE);
	    	 controllerDirection = REVERSE;
	    	 Output = -1*Output;
	    }	  
        //digitalWrite(InA1, LOW);  // CCW
        //digitalWrite(InB1, HIGH); // CCW
      }
      break;
    case 'o':                     // user should type "oo"
      digitalWrite(InA1, LOW);
      digitalWrite(InB1, LOW);
      speed_req = 0;
      break;
    default: 
      Serial.println("???");
    }
}

int digital_smooth(int value, int *data_array)  {    // remove signal noise
  static int ndx=0;                                                         
  static int count=0;                          
  static int total=0;                          
  total -= data_array[ndx];               
  data_array[ndx] = value;                
  total += data_array[ndx];               
  ndx = (ndx+1) % NUMREADINGS;                                
  if(count < NUMREADINGS) count++;
  return total/count;
}

/********************************************
 * Serial Communication functions / helpers
 ********************************************/

union {                // This Data structure lets
  byte asBytes[24];    // us take the byte array
  float asFloat[6];    // sent from processing and
}                      // easily convert it to a
foo;                   // float array

// getting float values from processing into the arduino
// was no small task.  the way this program does it is
// as follows:
//  * a float takes up 4 bytes.  in processing, convert
//    the array of floats we want to send, into an array
//    of bytes.
//  * send the bytes to the arduino
//  * use a data structure known as a union to convert
//    the array of bytes back into an array of floats

//  the bytes coming from the arduino follow the following
//  format:
//  0: 0=Manual, 1=Auto, else = ? error ?
//  1: 0=Direct, 1=Reverse, else = ? error ?
//  2-5: float setpoint
//  6-9: float input
//  10-13: float output  
//  14-17: float P_Param
//  18-21: float I_Param
//  22-245: float D_Param
void SerialReceive()
{
  // read the bytes sent from Processing
  int index=0;
  byte Auto_Man = -1;
  byte Direct_Reverse = -1;
  while(Serial.available()&&index<26)
  {
    if(index==0) Auto_Man = Serial.read();
    else if(index==1) Direct_Reverse = Serial.read();
    else foo.asBytes[index-2] = Serial.read();
    index++;
  } 
  
  // if the information we got was in the correct format, 
  // read it into the system
  if(index==26  && (Auto_Man==0 || Auto_Man==1)&& (Direct_Reverse==0 || Direct_Reverse==1))
  {
    Setpoint=double(foo.asFloat[0]);
    //Input=double(foo.asFloat[1]);       // * the user has the ability to send the 
                                          //   value of "Input"  in most cases (as 
                                          //   in this one) this is not needed.
    if(Auto_Man==0)                       // * only change the output if we are in 
    {                                     //   manual mode.  otherwise we'll get an
      Output=double(foo.asFloat[2]);      //   output blip, then the controller will 
    }                                     //   overwrite.
    
    double p, i, d;                       // * read in and set the controller tunings
    p = double(foo.asFloat[3]);           //
    i = double(foo.asFloat[4]);           //
    d = double(foo.asFloat[5]);           //
    myPID.SetTunings(p, i, d);            //
    
    if(Auto_Man==0) myPID.SetMode(MANUAL);// * set the controller mode
    else myPID.SetMode(AUTOMATIC);             //
    
    if(Direct_Reverse==0) myPID.SetControllerDirection(DIRECT);// * set the controller Direction
    else myPID.SetControllerDirection(REVERSE);          //
  }
  Serial.flush();                         // * clear any random data from the serial buffer
}

// unlike our tiny microprocessor, the processing ap
// has no problem converting strings into floats, so
// we can just send strings.  much easier than getting
// floats from processing to here no?
void SerialSend()
{
  Serial.print("PID ");
    Serial.print(Setpoint);   
    Serial.print(" ");
    Serial.print(Input);   
    Serial.print(" ");
    Serial.print(Output);   
    Serial.print(" ");
    Serial.print(myPID.GetKp());   
    Serial.print(" ");
    Serial.print(myPID.GetKi());   
    Serial.print(" ");
    Serial.print(myPID.GetKd());   
    Serial.print(" ");
    if(myPID.GetMode()==AUTOMATIC) Serial.print("Automatic");
    else Serial.print("Manual");  
    Serial.print(" ");
    if(myPID.GetDirection()==DIRECT) Serial.println("Direct");
  else Serial.println("Reverse");
}

